from rbx.gcp import PubSubPublisher


def notify_mdm(payload, project, topic, version):
    """Send an event to Google Pub/Sub to notify the change."""
    PubSubPublisher(
        project=project,
        topic=topic,
        version=version
    ).dispatch(payload)



import logging

from google.cloud.storage import Client

logger = logging.getLogger(__name__)


class Storage:
    """Provides functionality to upload/download content to/from GCS."""
    def __init__(self, project_id, bucket=None):
        self.client = Client()
        self.bucket = self.client.get_bucket(bucket or f'{project_id}.appspot.com')

    def download(self, location):
        blob = self.bucket.blob(location)
        if blob.exists():
            logger.debug(f'Downloading from storage: "{blob.name}"')
            content = blob.download_as_string()
            return content.decode('utf-8')

    def list_objects(self, prefix=None):
        """Return an iterator of all objects in the bucket matching the prefix.

        The objects are returned as bytes.
        """
        for blob in self.bucket.list_blobs(prefix=prefix.lstrip('/')):
            yield blob.download_as_string()

    def upload(self, content, location, content_type='text/plain'):
        blob = self.bucket.blob(location.lstrip('/'))
        logger.debug(f'Uploading to storage: "{blob.name}"')
        blob.upload_from_string(
            content.encode('utf-8'),
            content_type=content_type
        )



import logging
import warnings

from rq_scheduler import Scheduler as RQScheduler

from .jobs import Jobs

logger = logging.getLogger(__name__)


class Scheduler(RQScheduler):

    def __init__(self, queue, interval=60):
        """Initialise the Job Scheduler.

        Parameters:
            queue (rbx.queue.Queue):
                The queue scheduled jobs will be submitted to.
            interval (int):
                How often the scheduler checks for new jobs to add to the queue (in seconds, can
                be floating-point for more precision).
                Defaults to 60 seconds.

        """
        super().__init__(connection=queue.queue.connection, interval=interval)

    def register_death(self):
        """Extend what is executed when death is coming.

        More specifically, make sure all scheduled jobs are cancelled.
        They will be re-created upon re-birth.
        """
        super().register_death()

        for job in self.get_jobs():
            logger.debug('Removing {!r}'.format(job))
            self.cancel(job)

    def register_jobs_from_settings(self, schedules):
        """Given a set of scheduled jobs, register them."""
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning, module='rq')
            for job in Jobs.from_dict(schedules).values():
                logger.debug('Scheduling {!r}'.format(job))
                self.cron(
                    cron_string=job.schedule,
                    func=job.task,
                    args=job.args,
                    kwargs=job.kwargs,
                    meta={'scheduled': True},
                    timeout=job.timeout
                )



import warnings

from sqlalchemy import create_engine, text


class Database:
    """A SQLAlchemy Engine with useful convenience methods.

    A single connection should be created for a thread, so it can manage its own pool of
    connections more efficiently.
    """

    def __init__(self, db_engine_url, clear=False):
        """Create a connection engine with default connection pooling."""
        self.engine = create_engine(
            db_engine_url,
            max_overflow=2,
            pool_pre_ping=True,
            pool_recycle=1800,  # 30 minutes
            pool_size=5,
            pool_timeout=30
        )

    @property
    def dialect(self):
        return self.engine.dialect

    def connect(self, **kwargs):
        """Proxy to the Engine.connect() method.

        Returns:
            Connection: the connection object.
        """
        return self.engine.connect(**kwargs)

    def execute(self, *args, **kwargs):
        """Convenience method to execute a single statement or expression.

        The call arguments are passed on to the SqlAlchemy execute() method unaltered, all wrapped
        within a single connection transaction block.

        For full UTF-8 support with MySQL, the NAMES collation must be set prior to making the
        request. Failing to do this would not decode non-ascii characters properly.
        """
        with self.engine.connect() as connection:
            if self.dialect.name == 'mysql':
                connection.execute('SET NAMES utf8mb4 COLLATE utf8mb4_unicode_ci')
            return connection.execute(*args, **kwargs)

    def has_table(self, table_name):
        return self.engine.dialect.has_table(self.engine, table_name)

    def query(self, statement, connection=None, scalar=False):
        """Run a SQL statement and return the results.

        Parameters:
            statement (str):
                The raw SQL statement to execute.
            connection (Connection):
                When a Connection object is provided, use that and do not attempt to connect.
                This allows multiple statements to be executed using the same connection, or to
                use a transactional connection.
            scalar (bool):
                Fetch the first column of the first row.
        """
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', '.*Division by 0.*', category=Warning)
            if connection:
                cursor = connection.execute(text(statement))
            else:
                with self.engine.connect() as connection:
                    cursor = connection.execute(text(statement))

            return cursor.scalar() if scalar else cursor.fetchall()



import os
from pathlib import Path

RBX_ENV = os.getenv('RBX_ENV', 'dev')
RBX_PROJECT = f'{RBX_ENV}-platform-eu'

GOOGLE_CLOUD_PROJECT = os.getenv('GOOGLE_CLOUD_PROJECT')

MDM_PUBSUB_TOPIC = os.getenv('MDM_PUBSUB_TOPIC', 'platform-notifications')
MDM_PAYLOAD_VERSION = 2

AWS_BUCKET = os.getenv('AWS_BUCKET')
AWS_REGION = os.getenv('AWS_REGION')

PROJECT_ROOT = Path(__file__).resolve().parent.parent



import json
import logging

from datetime import datetime
from google.cloud import pubsub

logger = logging.getLogger(__name__)


def strip(message):
    """Strip whitespaces from the message payload."""
    payload = {
        k: v.strip() if type(v) is str else v
        for k, v in message['payload'].items()
    }
    message['payload'] = payload
    return message


class PubSubPublisher:
    """Publish messages to Pub/Sub."""
    def __init__(self, project, topic, version):
        self.project = project
        self.publisher = pubsub.PublisherClient()
        self.topic = topic
        self.version = version

    def _format(self, message, timestamp):
        """Format a message item into a Pub/Sub message."""
        return {
            'version': self.version,
            'timestamp': timestamp,
            'payload': message,
        }

    def _publish(self, message):
        """Publish to Pub/Sub.

        We wait for the result to ensure the publish operation has succeeded. The 30 seconds
        timeout is set in case something goes horribly wrong, which would lock the thread forever.

        Note that the Pub/Sub message must be sent as a bytestring.
        """
        logger.debug('Notifying "{}": <{}>'.format(self.topic, message))
        future = self.publisher.publish(
            f'projects/{self.project}/topics/{self.topic}',
            json.dumps(strip(message)).encode('utf-8')
        )
        return future.result(30)

    def dispatch(self, payload):
        """Dispatch events to be published.

        The Pub/Sub library takes care of batching behind the hook, so all we
        need to do is post the messages independently.
        """
        self.timestamp = datetime.utcnow().isoformat()

        if type(payload) in (list, tuple):
            messages = [self._format(message, timestamp=self.timestamp)
                        for message in payload]
        else:
            messages = [self._format(payload, timestamp=self.timestamp)]

        for message in messages:
            self._publish(message)



from enum import Enum, IntEnum
import logging
import logging.config
import os

import click
from google.cloud.logging_v2.handlers._helpers import get_request_data
from google.cloud.logging_v2.client import Client
from google.cloud.logging_v2.logger import Logger
from google.cloud.logging_v2.resource import Resource

ENABLE_CLOUD_LOGGING = 'ENABLE_CLOUD_LOGGING'


class CloudLoggingSeverity(IntEnum):
    NOTSET = 0
    DEBUG = 100
    INFO = 200
    WARNING = 400
    ERROR = 500
    CRITICAL = 600


class Service(Enum):
    AppEngine = 'gae_app'
    Dataflow = 'dataflow_step'


class CloudLoggingHandler(logging.Handler):
    """Logs to Google Cloud Logging."""

    def __init__(self, level=logging.DEBUG, name='rbx', service=Service.AppEngine):
        super().__init__(level=level)
        if self.cloud_logging_enabled:
            if not isinstance(service, Service):
                service = Service(service)
            self.writer = CloudLoggingWriter(name=name, service=service)

    @property
    def cloud_logging_enabled(self):
        return all([
            bool(os.getenv(ENABLE_CLOUD_LOGGING, False)),
            bool(os.getenv('GOOGLE_CLOUD_PROJECT', False)),
        ])

    def emit(self, record):
        if self.cloud_logging_enabled:
            if record.exc_info:
                message = logging.Formatter().formatException(record.exc_info)
            else:
                message = self.format(record)

            inferred_http, _ = get_request_data()
            http_request = getattr(record, 'http_request', inferred_http or {})

            entry = {
                'message': message,
                'python_logger': record.name,
                **http_request,
            }

            self.writer.write_entry(entry, severity=CloudLoggingSeverity[record.levelname])


class CloudLoggingWriter:
    """The Writer is the object that calls the Cloud Logging API."""

    def __init__(self, name, service):
        project = os.getenv('GOOGLE_CLOUD_PROJECT')
        client = Client(project=project)

        self.log_name = f'projects/{project}/logs/{name}'
        self.logger = Logger(self.log_name, client)

        labels = {'project_id': project}
        if service is Service.AppEngine:
            labels.update({
                'module_id': os.getenv('GAE_SERVICE', 'N/A'),
                'version_id': os.getenv('GAE_VERSION', 'N/A')
            })

        self.resource = Resource(type=service.value, labels=labels)

    def write_entry(self, message, severity=CloudLoggingSeverity.NOTSET):
        self.logger.log_struct(
            message,
            log_name=self.log_name,
            resource=self.resource,
            severity=severity
        )


class ColourFormatter(logging.Formatter):

    def format(self, record):
        message = super().format(record)
        colours = {
            logging.INFO: {'bold': True},
            logging.WARNING: {'fg': 'yellow'},
            logging.ERROR: {'fg': 'bright_red'},
            logging.CRITICAL: {'fg': 'bright_white', 'bg': 'red'},
        }
        try:
            message = click.style(message, **colours[record.levelno])
        except KeyError:
            pass
        return message



from dataclasses import asdict, dataclass, field, InitVar
import logging
import os
from unittest import mock

import google.auth.credentials
from google.cloud.firestore import Client

from ..settings import RBX_PROJECT

logger = logging.getLogger('rbx.auth')


class Keystore:
    """Manages API Keys stored in Google Cloud Firestore (Native Mode)."""

    def __init__(self):
        if os.getenv('GAE_ENV', '').startswith('standard'):
            db = Client(project=RBX_PROJECT)
        else:
            credentials = mock.Mock(spec=google.auth.credentials.Credentials)
            db = Client(project=os.getenv('GOOGLE_CLOUD_PROJECT'), credentials=credentials)

        self.collection = db.collection('api_keys')

    def _key(self, document):
        """Given a DocumentSnapshot, return the fully loaded Key object."""
        return Key(**{'keystore': self, 'key_id': document.id, **document.to_dict()})

    def create_key(self, email, name, campaigns=None, is_restricted=None, services=None):
        """Make a new API key.

        A key is associated with a user via her email address. If a key already exists for that
        user, the key is updated instead.
        """
        key = self.get_key(email=email)
        if not key:
            key = {
                'campaigns': campaigns or [],
                'email': email,
                'is_restricted': is_restricted if is_restricted is not None else True,
                'name': name,
                'services': services if services is not None else ['*.*'],
                'status': 'active',
            }

            _, ref = self.collection.add(key)
            key = self._key(ref.get())
        else:
            key.update(**{
                'campaigns': campaigns or key.campaigns,
                'is_restricted': is_restricted if is_restricted is not None else key.is_restricted,
                'name': name or key.name,
                'services': services or key.services,
                'status': 'active',  # the key is resuscitated
            })

        return key

    def get_key(self, key_id=None, email=None):
        """Retrieve a key by ID or email."""
        if key_id:
            document = self.collection.document(key_id).get()
            if document.exists:
                return self._key(document)

        if email:
            document = next(self.collection.where('email', '==', email).stream(), False)
            if document:
                return self._key(document)

    def update_key(self, key, attributes):
        """Set the new Key values in Firestore."""
        assert type(key) is Key, f'Excepted rbx.auth.Key, got {type(key)}'
        document = self.collection.document(key.key_id)
        document.set(attributes, merge=True)


@dataclass
class Key:
    key_id: InitVar[str]
    keystore: InitVar[Keystore]
    email: str
    name: str
    campaigns: list = field(default_factory=list)
    is_restricted: bool = True
    services: list = field(default_factory=['*.*'])
    status: str = 'active'

    def __post_init__(self, key_id, keystore):
        # These are defined as InitVar so that they are not part of the pickled data, and aren't
        # included in the to_dict() representation.
        self.key_id = key_id
        self.keystore = keystore

    def activate(self):
        self.update(status='active')

    def deactivate(self):
        self.update(status='inactive')

    def has_access(self, service, operation):
        """Check whether the service and operation are granted access by this Key."""
        if self.status != 'active':
            return False

        for grant in self.services:
            grant = dict(zip(
                ('service', 'operation'),
                grant.split('.')
            ))
            if grant['service'] == '*':
                return True
            elif grant['service'] == service and grant['operation'] == '*':
                return True
            elif grant['service'] == service and grant['operation'] == operation:
                return True

        return False

    def to_dict(self):
        return asdict(self)

    def update(self, **kwargs):
        for key, value in kwargs.items():
            if key in self.__annotations__.keys() and key not in ('key_id', 'keystore'):
                setattr(self, key, value)

        self.keystore.update_key(self, self.to_dict())


def fake_key(key_id, campaigns=None, email=None, is_restricted=True, name=None, services=None,
             status='active'):
    return Key(keystore=mock.Mock(spec_set=Keystore),
               key_id=key_id,
               campaigns=campaigns or [],
               email=email or 'john.doe@rip.com',
               is_restricted=is_restricted,
               name=name or 'John Doe',
               services=services if services is not None else ['*.*'],
               status=status)



import csv
import json
import sqlite3
import shutil
from itertools import groupby
from os import path, makedirs


def clean(value):
    """Strip trailing whitespaces and quotes."""
    return value.strip(' "')


class LocationPacker:
    """GeoIP Location Packer.

    Extracts data from the various database files, and generates a series of files that the Rig UI
    and the Platform can use.

    The database is the MaxMind GeoIP2 CSV Databases (City and Country).
    See https://dev.maxmind.com/geoip/geoip2/geoip2-city-country-csv-databases/
    """
    encoding = 'utf-8'

    def __init__(self, db_path, data_dir, version):
        self.db_path = path.normpath(db_path)
        self.version = version
        self.data_dir = path.join(path.normpath(data_dir), self.version)

        self.areas = []
        self.countries = {}  # This will hold the final JSON payload

    def __call__(self):
        # Wipe out the destination folder.
        if path.exists(self.data_dir):
            shutil.rmtree(self.data_dir)
        makedirs(path.join(self.data_dir, 'countries'))

        # Parse GeoIP database files.
        try:
            self.read_countries()
            self.read_cities()
        except (FileNotFoundError, TypeError):
            raise SystemExit

        # Unpack.
        self.unpack()

        # Generate SQLite3 database.
        self.create_database()

    def create_database(self):
        """Create the areas sqlite3 database."""

        # De-duplicated area labels
        self.areas = sorted(self.areas, key=lambda x: (x['country'], x['label'], x['region']))

        areas = []
        for item, _ in groupby(self.areas, lambda x: (x['country'], x['label'], x['region'])):
            areas.append({
                'country': item[0],
                'label': item[1],
                'region': item[2],
            })

        area_db = path.join(self.data_dir, 'area.db')
        with sqlite3.connect(area_db) as conn:
            c = conn.cursor()
            c.execute('CREATE TABLE labels (country text, label text, region text)')
            c.executemany(
                'INSERT INTO labels (country, label, region) VALUES (?, ?, ?)',
                [(area['country'], area['label'], area['region']) for area in areas]
            )
            conn.commit()

    def read_cities(self):
        """Read the region and city from the 'GeoIP2-City-Locations-en.csv' file.

         - Add region codes (subdivisions) to countries.
         - Set areas.
         - Must be read after countries.

        """
        database = path.join(self.db_path, 'GeoIP2-City-Locations-en.csv')
        with open(database, encoding=self.encoding) as fd:
            reader = csv.DictReader(fd)
            for row in reader:

                # Skip entries with no city.
                if not row['city_name']:
                    continue

                # Case 1: both sub-divisions are present; the region code is compiled from both
                # codes, separated by a dot (i.e.: '<sd1_code>.<sd2_code>'); the second division
                # name is used as the region name, with the first division in brackets.
                if row['subdivision_1_iso_code'] and row['subdivision_2_iso_code']:
                    region_code = '{}.{}'.format(
                        row['subdivision_1_iso_code'], row['subdivision_2_iso_code'])
                    region_name = '{} ({})'.format(
                        row['subdivision_2_name'], row['subdivision_1_name'])

                # Case 2: only the first sub-division is filled in; we just use that for region
                # code and name.
                elif row['subdivision_1_iso_code']:
                    region_code = row['subdivision_1_iso_code']
                    region_name = row['subdivision_1_name']

                # Case 3: no sub-divisions at all. We create a dummy 'Unspecified Region' region
                # with code 'N/A'.
                else:
                    region_code = 'N/A'
                    region_name = 'Unspecified Region'

                if region_code not in self.countries[row['country_iso_code']]['regions']:
                    self.countries[row['country_iso_code']]['regions'][region_code] = {
                        'name': region_name,
                        'cities': set([row['city_name']]),
                    }
                else:
                    self.countries[row['country_iso_code']]['regions'][
                        region_code
                    ]['cities'].add(row['city_name'])

                # Add each region to the list of areas.
                # Ignore the 'Unspecified Region' labels.
                if region_code != 'N/A':
                    self.areas.append({
                        'country': row['country_iso_code'],
                        'label': region_name,
                        'region': region_code
                    })

    def read_countries(self):
        """Read the country code and name from the 'GeoIP2-Country-Locations-en.csv' file.

         - Set initial countries.
         - Must be read before cities.

        """
        database = path.join(self.db_path, 'GeoIP2-Country-Locations-en.csv')
        with open(database, encoding=self.encoding) as fd:
            reader = csv.DictReader(fd)
            for row in reader:
                country = clean(row['country_iso_code'])
                if country:
                    self.countries[country] = {
                        'name': clean(row['country_name']),
                        'regions': {},
                    }

    def unpack(self):
        """Convert GeoIP locations to JSON files."""

        # Output country list (geo.json)
        # Note that countries are grouped using their names, not their codes.
        # ------------------------------
        # Format:
        #   {
        #     "categories": [
        #       {
        #         "sub": [
        #           {
        #             "value": "<country_code>",
        #             "label": "<country_name>"
        #           },
        #           ...
        #         ],
        #         "label": "<country_name>[:1]",
        #         "value": "<country_name>[:1]"
        #       },
        #       ...
        #     ]
        #   }
        #

        # Order countries dictionary by country name
        sorted_by_name = sorted(self.countries.items(), key=lambda kv: kv[1]['name'])
        self.countries = dict(sorted_by_name)

        categories = []
        for code, value in self.countries.items():
            letter = value['name'][:1]
            try:
                category = [cat for cat in categories if cat['value'] == letter][0]
            except IndexError:
                categories.append({
                    'sub': [{
                        'label': value['name'],
                        'value': code
                    }],
                    'label': letter,
                    'value': letter,
                })
            else:
                categories[categories.index(category)]['sub'].append({
                    'label': value['name'],
                    'value': code
                })

        output = {'categories': categories}

        with open(path.join(self.data_dir, 'geo.json'), mode='w', encoding='utf-8') as fd:
            fd.write(json.dumps(output, indent=2))

        # Create country files (countries/<country_code>.json)
        # ------------------------------
        # Format:
        #   {
        #     "categories": [
        #       {
        #         "country": "<country_code>",
        #         "value": "<region_code>",
        #         "label": "<region_name>",
        #         "sub": [
        #           {
        #             "value": "<city>",
        #             "label": "<city>"
        #           },
        #           ...
        #         ]
        #       },
        #       ...
        #     ]
        #   }
        #
        for country_code, country in self.countries.items():
            output = {
                'categories': [
                    {
                        'country': country_code,
                        'label': region['name'],
                        'value': code,
                        'sub': [
                            {
                                'label': city,
                                'value': city
                            }
                            for city in sorted(region['cities'])
                        ],
                    }
                    for code, region in sorted(country['regions'].items())
                ]
            }

            country_file = path.join(self.data_dir, 'countries', country_code + '.json')
            with open(country_file, mode='w', encoding='utf-8') as fd:
                fd.write(json.dumps(output))



from io import BytesIO


class BaseQuery:
    def __init__(self, table_name, key_field, schema, prefix=None, suffix='staging'):
        self._sql = None
        self.fields = [field.name for field in schema]
        self.key_field = key_field
        self.table = '{prefix}{table}'.format(
            prefix='{}.'.format(prefix) if prefix else '',
            table=table_name)
        self.staging = '{prefix}{table}_{suffix}'.format(
            prefix='{}.'.format(prefix) if prefix else '',
            table=table_name,
            suffix=suffix)


class BigQueryUpdateQuery(BaseQuery):
    """Encapsulate the generation of SQL UPDATE queries for BigQuery.

    The query is used to update a dimension's table according to a staging table.
    """
    def __init__(self, *args, **kwargs):
        kwargs['prefix'] = 'dimensions'
        super().__init__(*args, **kwargs)

    @property
    def sql(self):
        """Generate the following SQL:

        UPDATE dimensions.<dimension> AS D
        SET D.<key> = S.<key>, D.<field> = S.<field>, ...
        FROM (
          -- Select all rows from both tables;
          -- All the columns in the rows which exist in "staging" and not in
          -- <dimension> will be NULL.
          SELECT staging.* FROM dimensions.<dimension>_staging AS staging
          LEFT OUTER JOIN dimensions.<dimension> AS dimension
          ON dimension.<key> = staging.<key>
          -- Filter out NULL rows, keeping all existing rows which must be
          -- updated.
          WHERE dimension.<key> IS NOT NULL
        ) S
        WHERE D.<key> = S.<key>;

        """
        if self._sql is None:
            field_set = ', '.join([
                'D.{field} = S.{field}'.format(field=field)
                for field in self.fields
            ])
            self._sql = (
                'UPDATE {table} AS D '
                'SET {field_set} '
                'FROM ( '
                'SELECT staging.* FROM {staging} AS staging '
                'LEFT OUTER JOIN {table} AS dimension '
                'ON dimension.{key} = staging.{key} '
                'WHERE dimension.{key} IS NOT NULL '
                ') S WHERE D.{key} = S.{key}'
            ).format(
                table=self.table,
                field_set=field_set,
                staging=self.staging,
                key=self.key_field
            )

        return self._sql


class InsertQuery(BaseQuery):
    """Encapsulate the generation of SQL INSERT queries for BigQuery and CloudSQL.

    The query is used to insert new rows from a staging table to a dimension's table.
    """
    @property
    def sql(self):
        """Generate the following SQL:

        INSERT INTO <dimension> (
          <key>, <field>, ...
        )
        -- Select all rows from both tables;
        -- All the columns in the rows which exist in "staging" and not in
        -- "dimension" will be NULL.
        SELECT staging.* FROM <dimension>_staging AS staging
        LEFT OUTER JOIN <dimension> AS dimension
        ON dimension.<key> = staging.<key>
        -- Filter only NULL rows, which are the new rows.
        WHERE dimension.<key> IS NULL;

        """
        if self._sql is None:
            field_set = ', '.join(self.fields)
            self._sql = (
                'INSERT INTO {table} ( {field_set} ) '
                'SELECT staging.* FROM {staging} AS staging '
                'LEFT OUTER JOIN {table} AS dimension '
                'ON dimension.{key} = staging.{key} '
                'WHERE dimension.{key} IS NULL'
            ).format(
                table=self.table,
                field_set=field_set,
                staging=self.staging,
                key=self.key_field
            )

        return self._sql


class CloudSQLUpdateQuery(BaseQuery):
    """Encapsulate the generation of SQL UPDATE queries for CloudSQL.

    The query is used to update a dimension's table according to a staging table.
    """
    @property
    def sql(self):
        """Generate the following SQL:

        UPDATE <dimension> as D, <dimension>_staging as S
        SET D.<key> = S.<key>, D.<field> = S.<field>, ...
        WHERE D.<key> = S.<key>;

        """
        if self._sql is None:
            field_set = ', '.join([
                'D.{field} = S.{field}'.format(field=field)
                for field in self.fields
            ])
            self._sql = (
                'UPDATE {table} AS D, {staging} AS S '
                'SET {field_set} '
                'WHERE D.{key} = S.{key}'
            ).format(
                table=self.table,
                field_set=field_set,
                staging=self.staging,
                key=self.key_field
            )

        return self._sql


class LoadQuery:
    """Encapsulate the generation of payload data for BigQuery loads.
    The payload is formatted using the CSV export format.
    """
    source_format = 'CSV'

    def __init__(self, payload, schema):
        self.fields = [field.name for field in schema]
        self.stream = BytesIO()
        if type(payload) is list:
            for item in payload:
                self._write(item)
        else:
            self._write(payload)

    @property
    def data(self):
        """Reset and return the stream object."""
        self.stream.seek(0)
        return self.stream

    def _write(self, line):
        """Write a CSV line using the same order as the schema fields.

        Some of the field values may be UTF-8 encoded already, and some not.
        We also need to keep the file content UTF-8 encoded in order for
        Big Query to load data correctly.
        We therefore need to decode UTF-8 encoded bytes, convert integers and
        booleans to unicodes, and encode the final CSV output.

        Also note that as we use the default double-quotes to enclose fields,
        a double-quote appearing inside a field must be escaped by preceding
        it with another double quote.
        """
        values = []
        for field in self.fields:
            value = line[field]
            if value is None:
                value = ''
            elif type(value) in (float, int, bool):
                value = str(value).lower()
            elif type(value) is bytes:
                value = '"{}"'.format(
                    value.decode('utf-8').replace('"', '""'))
            else:
                value = '"{}"'.format(value.replace('"', '""'))
            values.append(value)

        values = ','.join(values).encode('utf-8')
        self.stream.write(values)
        self.stream.write(b'\n')



import json
import logging
from os import getenv, path
import shelve

import arrow

from ..exceptions import ClientError
from . import Client, HttpAuth

logger = logging.getLogger(__name__)


class NovatiqClient(Client):
    """Client for Novatiq CreatOR API.

    The API documentation can be found on the Platform wiki:
    https://github.com/rockabox/rbx/wiki/Novatiq

    """
    AUTH_PATH = 'auth'
    CATEGORIES = {
        'Family & Parenting': 'Family',
        'Food & Drink': 'Food and Drink',
        'Law, Government & Politics': 'Law & Govt',
        'News': 'Lifestyle',
        'Style & Fashion': 'Fashion',
        'Religion & Spirituality': 'Lifestyle',
        'Uncategorized': 'Lifestyle',
        'Non-Standard Content': 'Lifestyle',
        'Illegal Content': 'Adult Content',
    }
    ENDPOINT = 'https://cpdemo.awsdev.smartpipesolutions.com/api/v1/'
    LIVE_ENDPOINT = 'https://access.cp.novatiq.com/api/v1/'
    TIME_FORMAT = 'YYYY-MM-DD'
    TOKEN = 'accessToken'

    def __init__(self, destination='/tmp'):
        super().__init__()
        if getenv('RBX_ENVIRONMENT', 'dev') == 'live':
            self.ENDPOINT = self.LIVE_ENDPOINT

        self.shelf = path.join(destination, 'novatiq.segments')

    @property
    def auth(self):
        """Novatiq uses Digest Authentication."""
        return HttpAuth(self.token, key='Authorization')

    def create_campaign(self, brand, category, country_code, name, segments, start_date, end_date):
        """Create a campaign.

        If the campaign name already exists, we fetch that campaign to return the existing campaign
        ID. Technically this should never happen, except possibly in a development setting.
        """
        try:
            response = self.request('post', 'campaigns', data=[{
                'name': name,
                'clientName': brand,
                'countryIso': country_code,
                'campaignCategory': self.CATEGORIES.get(category, category),
                'startDate': arrow.get(start_date).format(self.TIME_FORMAT),
                'endDate': arrow.get(end_date).format(self.TIME_FORMAT),
                'segments': segments,
            }])
            return response[0]['id']
        except ClientError as e:
            error = json.loads(str(e))
            if error['errorInfo']['errorCode'] == '1200002':
                # Code '1200002' means 'Given name already exists'
                return self.get_campaign(name=name)['id']
            else:
                raise

    def get_campaign(self, campaign_id=None, name=None):
        """Retrieve the campaign given either its ID or its name."""
        response = self.request('get', 'campaigns')
        field = 'id' if campaign_id else 'name'
        value = campaign_id or name
        try:
            campaign = [row for row in response if row.get(field) == value][0]
        except IndexError:
            return None

        return {
            'id': campaign['id'],
            'segments': [segment['id'] for segment in campaign['segments']],
        }

    def get_message(self, response):
        """Extract the message from the requests response object."""
        return self.get_response(response).get('errorInfo', {}).get('errorMessage')

    def get_segments(self, country_code=None):
        """Get segments from local cache."""
        with shelve.open(self.shelf) as shelf:
            try:
                segments = shelf['segments']
            except KeyError:
                # The cache is empty
                segments = []

        segment_list = []
        for segment in segments:
            if country_code and country_code not in [segment['country'], 'ZZ']:
                continue
            segment_list.append(segment)

        return segment_list

    def list_segments(self):
        """List all segments available.

        According to the documentation, there is no pagination on this endpoint, so we expect the
        full list to be returned with a single call.

        See: https://cpdemo.awsdev.smartpipesolutions.com/api/docs/index.html#/segment/getSegments
        """
        response = self.request('get', 'segments')

        return [{
            'country': segment['countryIso'],
            'cpm': segment['segmentPrice'],
            'currency': 'TRY',
            'expression': segment['expression'],
            'id': segment['id'],
            'name': segment['name'],
        } for segment in response]

    def pause_campaign(self, campaign_id):
        """Pause a campaign."""
        response = self.request('post', f'campaigns/{campaign_id}/pause')
        return response['id']

    def start_campaign(self, campaign_id):
        """Start a campaign."""
        response = self.request('post', f'campaigns/{campaign_id}/start')
        return response['id']

    def store_segments(self):
        """Retrieve the list of segments, and then cache the results on a shelf.

        As no caching mechanism exists on the API, we simply override the content of the cache
        with the results from the list_segments() call.
        """
        segments = self.list_segments()
        with shelve.open(self.shelf) as shelf:
            shelf['segments'] = segments

    def update_campaign(self, campaign_id, name=None, segments=None, end_date=None):
        """Create a campaign.

        Only the name, segments, and end date can be changed.
        When providing a list of segments, the full list of segments must be provided. All missing
        ones will be removed from the campaign.
        """
        payload = {}

        if name is not None:
            payload['name'] = name

        if end_date is not None:
            payload['endDate'] = arrow.get(end_date).format(self.TIME_FORMAT)

        if segments:
            campaign = self.get_campaign(campaign_id=campaign_id)
            if campaign is not None:
                attached = set(campaign['segments'])
                segments_to_add = list(set(segments) - attached)
                if segments_to_add:
                    payload['segmentsToAdd'] = segments_to_add
                segments_to_remove = list(attached - set(segments))
                if segments_to_remove:
                    payload['segmentsToRemove'] = segments_to_remove

        if payload:
            response = self.request('post', f'campaigns/{campaign_id}', data=payload)
            return response['id']

        return campaign_id



from invoke import Config, Program

from .tasks import deploy


class BuildtoolsConfig(Config):
    prefix = 'rbx'


program = Program(config_class=BuildtoolsConfig, namespace=deploy.ns, version='2.15.1')



from collections import namedtuple
from enum import Enum


Geocode = namedtuple('Geocode', ['location', 'location_type', 'viewport'])
Location = namedtuple('Location', [
    'city', 'country', 'formatted_address', 'place_id', 'postcode', 'region'
])
Point = namedtuple('Point', ['latitude', 'longitude'])


class Viewport(namedtuple('Viewport', ['southwest', 'northeast'])):
    def __contains__(self, point):
        """A Point is in the Viewport when its lat/long coordinates are within the SW/NE bounding
        box.

        Latitude
               ^
               |              NE
               +-----+-------+
               |     |       |
               x--------P    |
               |     |  |    |
               +-----+--|----+
               |   SW|  |    |
               o-----+--x----+----------> Longitude

        """
        if type(point) is Point:
            return all([
                point.latitude >= self.southwest.latitude,
                point.latitude <= self.northeast.latitude,
                point.longitude >= self.southwest.longitude,
                point.longitude <= self.northeast.longitude,
            ])

        return False


class Geotype(Enum):
    ROOFTOP = (
        'Indicates that the returned result is a precise geocode for which we have location '
        'information accurate down to street address precision.'
    )
    RANGE_INTERPOLATED = (
        'Indicates that the returned result reflects an approximation (usually on a road) '
        'interpolated between two precise points (such as intersections). Interpolated results '
        'are generally returned when rooftop geocodes are unavailable for a street address.'
    )
    GEOMETRIC_CENTER = (
        'Indicates that the returned result is the geometric center of a result such as a '
        'polyline (for example, a street) or polygon (region).'
    )
    APPROXIMATE = (
        'Indicates that the returned result is approximate.'
    )
    MANUAL = (
        'Indicates that the Coordinates were added manually.'
    )
    MISSING = (
        'Indicates that a result would not be found.'
    )
    MULTIPLE = (
        'Indicates that the Coordinates are an alternative location for an existing location.'
    )



import json
import logging

from google.cloud.bigquery import Client, LoadJobConfig, QueryJobConfig, Table
from google.cloud.exceptions import BadRequest, NotFound

from ..exceptions import FatalException
from .query import InsertQuery, LoadQuery, BigQueryUpdateQuery

logger = logging.getLogger(__name__)

TIMEOUT_MS = 20000


class BigQuery:

    def __init__(self, project):
        self.project = project

        self._client = None

    @property
    def client(self):
        if not self._client:
            self._client = Client(project=self.project)
        return self._client

    def create_table(self, dataset, table_name, schema, recreate=False):
        """Create the table in BigQuery.

        If the table already exists, do nothing.
        Unless recreate is set to True, in which case we drop the table first.
        """
        table_id = f'{self.project}.{dataset}.{table_name}'

        try:
            table = self.client.get_table(table_id)
        except NotFound:
            logger.info(f'Creating "{table_name}" table (BigQuery) ...')
            table = self.client.create_table(Table(table_id, schema=schema))
        else:
            # Table already exists
            if recreate:
                logger.info(f'Re-creating "{table_name}" table (BigQuery) ...')
                self.client.delete_table(table_id)
                table = self.client.create_table(Table(table_id, schema=schema))

        return table

    def insert(self, table_name, key_field, schema):
        """Run an INSERT query."""
        query = InsertQuery(table_name=table_name, key_field=key_field, schema=schema,
                            prefix='dimensions')
        logger.info(f'Inserting into "{query.table}" table (BigQuery) ...')

        job_config = QueryJobConfig()
        query_job = self.client.query(query.sql, job_config=job_config)
        query_job.result(timeout=TIMEOUT_MS)

    def load(self, dataset, table_name, schema, payload):
        """Load a payload to BigQuery."""
        logger.info(f'Loading {len(payload)} records to "{table_name}" table (BigQuery) ...')
        query = LoadQuery(payload, schema=schema)

        table_id = f'{self.project}.{dataset}.{table_name}'
        table = self.client.get_table(table_id)

        job_config = LoadJobConfig()
        job_config.schema = schema
        job_config.source_format = query.source_format
        job = self.client.load_table_from_file(query.data, table, job_config=job_config)
        try:
            job.result()
        except BadRequest:
            # The error will be set in the job.errors property, so no need to log it
            # again. We do add the requested payload to help troubleshooting.
            job.errors.append({'payload': query.data.read().decode('utf-8')})

        if job.errors:
            raise FatalException(json.dumps(job.errors))

        return table

    def update(self, table_name, key_field, schema):
        """Run an UPDATE query."""
        query = BigQueryUpdateQuery(table_name=table_name, key_field=key_field, schema=schema)
        logger.info(f'Updating "{query.table}" table (BigQuery) ...')

        job_config = QueryJobConfig()
        query_job = self.client.query(query.sql, job_config=job_config)
        query_job.result(timeout=TIMEOUT_MS)

    def update_table(self, dataset, table_name, schema):
        """Update a table schema in BigQuery.

        BigQuery schema updates only support adding columns, or editing them.
        Removing columns is not supported.
        """
        logger.info(f'Updating "{table_name}" table (BigQuery) ...')
        table_id = f'{self.project}.{dataset}.{table_name}'
        table = self.client.get_table(table_id)

        if len(table.schema) <= len(schema):
            table.schema = schema
            self.client.update_table(table, ['schema'])



from collections import namedtuple
from collections.abc import Mapping
import warnings

# Ignore Google Cloud warnings about using end user credentials.
warnings.filterwarnings('ignore', 'Your application has authenticated using end user credentials')


def namedtuple_with_defaults(typename, field_names, default_values=()):
    """Currently defaults for namedtuples are not supported right out the box.
    Python 3.7 does have support for this however until we are at that point this method
    serves the purpose.

    This method sets None to all specified `field_names`. However it does support specific
    defaults. e.g. default_values={name='test'}.
    """
    T = namedtuple(typename, field_names)
    T.__new__.__defaults__ = (None,) * len(T._fields)
    if isinstance(default_values, Mapping):
        prototype = T(**default_values)
    else:
        prototype = T(*default_values)
    T.__new__.__defaults__ = tuple(prototype)
    return T


Aggregate = namedtuple('Aggregate', ['name', 'schema', 'table'])

Dimension = namedtuple_with_defaults(
    'Dimension',
    ['name', 'key', 'ancestor', 'model_based', 'key_field', 'id_fields', 'index_fields',
     'relatives', 'schema', 'tables'],
    default_values={'model_based': False, 'index_fields': ()}
)

Fact = namedtuple('Fact', ['name', 'schema'])

__all__ = ['Aggregate', 'Dimension', 'Fact']



from copy import deepcopy as copy
import logging
import re

from sqlalchemy import text, MetaData, Table

from .query import CloudSQLUpdateQuery, InsertQuery

logger = logging.getLogger(__name__)


def _format_for_mysql(item):
    """Format payload data in a MySQL-friendly format.

    This involves:

      - Stripping any UTC ZONED ISO Format field from its timezone value (Z|+00:00).
        (The field is expected to be in ISO 8601 format.)

      - Decoding all previously-encoded bytes back to unicode.

    """

    # Regex to determine if a value is an ISO Format timestamp with zone.
    # Supports both +00:00 and Z. e.g.
    # 2019-04-04T15:13:53.801759+00:00
    # 2019-04-04T15:13:53.801759Z
    iso_format_regex = re.compile(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+(?:Z|\+00:00)')

    for key, value in item.items():
        if key.endswith('_name') and type(value) is bytes:
            item[key] = value.decode('utf-8')

        if value is not None and iso_format_regex.match(str(value)) is not None:
            # Handle the Z case.
            if value.endswith('Z'):
                item[key] = value[:-1]
                continue

            # Handle the +00:00 case.
            item[key] = value.split('+')[0]

    return item


class CloudSQL:
    def __init__(self, db):
        self.db = db
        self.metadata = MetaData()

    def create_table(self, table, recreate=False):
        """Create the table in CloudSQL.

        By default, will not attempt to recreate the table already present.
        Unless recreate is set to True.
        """
        assert type(table) is Table

        logger.info('{} "{}" table (CloudSQL) ...'.format(
            {True: 'Re-creating', False: 'Creating'}[recreate],
            table.name))

        self.metadata.create_all(self.db.engine, tables=[table])
        if recreate:
            self.db.execute(text('TRUNCATE TABLE {}'.format(table.name)))

    def insert(self, table_name, key_field, schema):
        """Run an INSERT query."""
        query = InsertQuery(table_name=table_name, key_field=key_field, schema=schema)
        logger.info('Inserting into "{}" table (CloudSQL) ...'.format(table_name))

        self.db.execute(text(query.sql))

    def load(self, table, payload):
        """Load a payload to CloudSQL.

        MySQL doesn't support custom timezones in the TIMESTAMP type, so we
        must manually remove all of the timezones from the 'timestamp' field.
        """
        assert type(table) is Table

        logger.info('Loading {} records to "{}" table (CloudSQL) ...'.format(
            len(payload), table.name))

        statement = table.insert()
        data = copy(payload)
        data[:] = list(map(_format_for_mysql, data))

        self.db.execute(statement, data)

    def update(self, table_name, key_field, schema):
        """Run an UPDATE query."""
        query = CloudSQLUpdateQuery(table_name=table_name, key_field=key_field, schema=schema)
        logger.debug('Updating "{}" table (CloudSQL) ...'.format(query.table))

        self.db.execute(text(query.sql))



from contextlib import contextmanager
from queue import Queue, Empty
import logging
import subprocess
from time import sleep
import threading

logger = logging.getLogger(__name__)


class CloudSQLProxy(threading.Thread):
    def __init__(self, command):
        self.command = command
        self.queue = Queue()
        threading.Thread.__init__(self)

    def run(self):
        self.process = subprocess.Popen(self.command.split(),
                                        shell=False,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE)
        # The cloud_sql_proxy CLI write to stderr
        for line in iter(self.process.stderr.readline, b''):
            self.queue.put(line)
        self.process.stdout.close()


@contextmanager
def cloud_sql_proxy(connection_name):
    """A context manager to wrap a CLI call with a Cloud SQL Proxy connection.

    The execution will be blocked until the connection is ready to accept connections.
    The connection will be closed on exit.

    Use as a context manager:
    >>> with cloud_sql_proxy(connection_name='project:region:instance'):
    >>>     do_your_thing()

    """
    command = f'./cloud_sql_proxy -instances={connection_name} -dir=./cloudsql -verbose=false'
    proxy = CloudSQLProxy(command=command)
    proxy.start()

    try:
        while True:
            try:
                line = proxy.queue.get_nowait().decode('utf-8').strip()
                if line:
                    logger.debug(line)
                if 'ready' in line.lower():
                    break  # Ready
            except Empty:
                pass  # Nothing yet
            sleep(2)
        yield proxy

    finally:
        proxy.process.terminate()
        proxy.join()


__all__ = ['cloud_sql_proxy']



"""Module to manage EDS schema configurations.

Configurations are drawn from a local schema configuration file.
The file must be named 'schema.yaml', and be accessible at the root of the project.
the file must be formatted using YAML.

To examine the current configuration from a Python shell:

    >>> from rbx.eds.config import schemas
    >>> schema.DIMENSIONS

"""
from collections.abc import Mapping
import os

from google.cloud.bigquery import SchemaField
from sqlalchemy import Column, MetaData, Table, Boolean, Float, Integer, BINARY, String
from sqlalchemy.dialects.mysql import DATETIME

import yaml

from . import Aggregate, Dimension, Fact

types = {
    'BINARY': BINARY(20),
    'BOOLEAN': Boolean,
    'INTEGER': Integer,
    'FLOAT': Float,
    'STRING': String,
    'TIMESTAMP': DATETIME(fsp=6),
}


def _table_for_aggregate(aggregate, metadata):
    """The database metadata using the SQLAlchemy representation.

    An aggregate table may have multiple Primary Keys.
    """
    fields = [
        Column(
            field['name'],
            types[field['type']](field.get('length', 255))
            if field['type'] == 'STRING' else types[field['type']],
            autoincrement=field.get('autoincrement', False),
            default=field.get('default', None),
            index=field.get('index', False),
            nullable=field.get('nullable', False),
            primary_key=field.get('primary_key', False),
            unique=field.get('unique', False),
        )
        for field in aggregate['schema']
    ]

    return Table(aggregate['name'], metadata, *fields, mysql_charset='utf8mb4')


def _tables_for_dimension(dimension, metadata):
    """The database metadata using the SQLAlchemy representation.

    The first field in the ordered list of fields is always assumed to be
    the Primary Key. This field is always an Integer.

    The meta will always include 2 tables, one for the table, and one for its staging table.
    """
    name = dimension['key']
    schema = dimension['schema']
    primary_key = schema[0]

    fields = [Column(primary_key['name'], Integer, primary_key=True, autoincrement=False)]
    fields_staging = [Column(primary_key['name'], Integer, primary_key=True, autoincrement=False)]

    secondary_fields = [field for field in schema if field['name'] != primary_key['name']]
    for field in secondary_fields:
        fields.append(Column(
            field['name'],
            types[field['type']](field.get('length', 255))
            if field['type'] == 'STRING' else types[field['type']],
            index=field['name'] in dimension.get('index_fields', []),
        ))
        fields_staging.append(Column(
            field['name'],
            types[field['type']](field.get('length', 255))
            if field['type'] == 'STRING' else types[field['type']],
            index=field['name'] in dimension.get('index_fields', []),
        ))

    return {
        name: Table(name, metadata, *fields, mysql_charset='utf8mb4'),
        name + '_staging': Table(name + '_staging', metadata, *fields_staging,
                                 mysql_charset='utf8mb4'),
    }


def _parse_yaml_config(data, metadata):
    """Parse the YAML configuration data into a dictionary.

    For example:

        dimensions:
          -
            name: string
            key: string

        facts:
          -
            name: string

    Becomes:

        {
            'DIMENSIONS': {
                'key': Dimension(name='string')
            },
            'FACTS': {
                'name': Fact(name='string')
            },
            'PURE_DIMENSIONS': {}
        }

    Dimensions are pure when they have a schema and are model-based.
    """
    values = {}
    if not data:
        return values
    try:
        content = yaml.load(data, Loader=yaml.FullLoader)
    except yaml.parser.ParserError:
        return values

    if 'aggregates' in content:
        values['AGGREGATES'] = {}
        for aggregate in content['aggregates']:
            aggregate['table'] = _table_for_aggregate(aggregate, metadata)
            aggregate['schema'][:] = [
                SchemaField(field['name'], field['type'],
                            'NULLABLE' if field.get('nullable', False) else 'REQUIRED')
                for field in aggregate['schema']
            ]
            values['AGGREGATES'][aggregate['name']] = Aggregate(**aggregate)

    if 'dimensions' in content:
        values['DIMENSIONS'] = {}
        for dimension in content['dimensions']:
            if 'key' in dimension:
                if 'schema' in dimension:
                    dimension['tables'] = _tables_for_dimension(dimension, metadata)
                    dimension['schema'][:] = [
                        SchemaField(field['name'], field['type'], field.get('mode', 'NULLABLE'))
                        for field in dimension['schema']
                    ]
                values['DIMENSIONS'][dimension['key']] = Dimension(**dimension)

        values['PURE_DIMENSIONS'] = {}
        for key, value in values['DIMENSIONS'].items():
            if value.model_based and value.schema:
                values['PURE_DIMENSIONS'][key] = value

    if 'facts' in content:
        values['FACTS'] = {}
        for fact in content['facts']:
            fact['schema'][:] = [
                SchemaField(field['name'], field['type'],
                            'REQUIRED' if field['required'] else 'NULLABLE')
                for field in fact['schema']
            ]
            values['FACTS'][fact['name']] = Fact(**fact)

    # All other sections are returned without further processing
    values.update({
        key.upper(): value
        for key, value in content.items()
        if key not in ('aggregates', 'dimensions', 'facts')
    })

    return values


def _configure(metadata):
    """This function loads the schema configuration from the YAML file."""
    defaultpath = os.path.join(os.path.dirname('__FILE__'), 'schema.yaml')
    filepath = os.getenv('SCHEMA_YAML_PATH', defaultpath)
    with open(filepath, 'rb') as fd:
        content = fd.read()

    return _parse_yaml_config(content, metadata)


class Schema(Mapping):
    """Lazy loader of schema data."""
    def __init__(self, metadata):
        self._data = None
        self.metadata = metadata

    @property
    def data(self):
        if self._data is None:
            self._data = _configure(metadata=self.metadata)
        return self._data

    def __getitem__(self, key):
        return self.data.get(key)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getattr__(self, name):
        return self[name]


schema = Schema(metadata=MetaData())

__all__ = ['schema']



import cv2
import numpy as np


def px_to_int(pixel_value):
    """Takes a pixel string in the format '123px' and returns the int number.

    Parameters:
        pixel_value (str):
            The pixel value, written in the format '123px'.

    Returns:
        int:
            The number extracted from the pixel value.
    """
    if not pixel_value:
        return 0
    elif isinstance(pixel_value, str):
        pixel_value = int(pixel_value.strip("px"))
    else:
        pixel_value = int(pixel_value)
    return pixel_value


def make_white_image(height, width):
    """Generate a white background of appropriate height/width and return.

    This white background also has an alpha channel added for easier merging with
    other images with alpha channels.

    Parameters:
        height (int):
            The height of the image to generate, in pixels.
        width (int):
            The width of the image to generate, in pixels.

    Returns:
        ndarray:
            An opencv compatible representation of the white image.
    """
    white_background = np.zeros([height, width, 3], dtype=np.uint8)
    white_background.fill(255)
    b_channel, g_channel, r_channel = cv2.split(white_background)
    alpha_channel = np.ones(b_channel.shape, dtype=b_channel.dtype) * 255
    white_background = cv2.merge((b_channel, g_channel, r_channel, alpha_channel))

    return white_background


def make_transparent_image(height, width):
    """Generate a transparent background of appropriate height/width and return.

    Parameters:
        height (int):
            The height of the image to generate, in pixels.
        width (int):
            The width of the image to generate, in pixels.

    Returns:
        ndarray:
            An opencv compatible representation of the transparent image.
    """
    transparent_background = np.zeros([height, width, 4], dtype=np.uint8)
    return transparent_background


def flatten(data):
    """Take a nested list and return a generator that will output each element, in order.

    Args:
        data (list):
            The list of nested lists.

    Returns:
        generator:
            A generator that will return each number in the flat list.
    """
    for entry in data:
        if isinstance(entry, list):
            yield from flatten(entry)
        else:
            yield entry



import logging

from google.api_core.retry import Retry
from google.api_core.exceptions import AlreadyExists, GoogleAPIError
from google.cloud import tasks_v2

logger = logging.getLogger(__name__)


class Queue:
    """Utility class to create queues and tasks in Google Cloud Tasks."""

    def __init__(self, project_id, location, queue_name, service='default'):
        """Initialise the Google Cloud Queue.

        Parameters:
            project_id (str):
                The Google Cloud Project ID.
            location (str):
                The location code of the Queue.
                This must match the location of the App Engine application.
            queue_name (str):
                The name of the queue.
            service:
                The default App Engine service that the task will be sent to.
                By default, tasks are sent to the 'default' service.

        """
        self.client = tasks_v2.CloudTasksClient()
        self.location = f'projects/{project_id}/locations/{location}'
        self.service = service

        name = self.client.queue_path(project_id, location, queue_name)
        self.queue = tasks_v2.Queue(name=name)

    @property
    def retry(self):
        """Exponential retry decorator.

        This class is a decorator used to add exponential back-off retry behavior to the
        create_task RPC call.

        We implement it here as a property of the queue, first to always use the same retry logic,
        and second to allow us to mock it more easily in tests.
        """
        return Retry(initial=.1, maximum=60, multiplier=1.3, deadline=600)

    def create(self):
        try:
            self.client.create_queue(parent=self.location, queue=self.queue)
        except AlreadyExists:
            logger.debug(f'Queue "{self.queue.name}" already exists.')
        except GoogleAPIError as e:
            logger.exception(e)

    def enqueue(self, uri, headers=None, method='POST', payload=None, service=None):
        """Create a task in the queue.

        This will create an App Engine HTTP request. The task will be delivered to the App Engine
        app which belongs to the same project as the queue.

        Parameters:
            uri (str):
                The url that the task will be sent to.
            headers (dict):
                Optional additional HTTP request headers.
                This map contains the header field names and values.
            method (str):
                The HTTP method to use for the request. Can be either GET or POST.
                The default is POST.
            payload (bytes):
                HTTP request body.
                A request body is allowed only if the HTTP method is POST.
                The payload must be encoded.
            service:
                The App Engine service that the task will be sent to.
                By default, the task is sent to the service implemented by the Queue.

        Raises:
            TypeError: If the payload parameter is not a bytes type.

        """
        http_method = {
            'GET': tasks_v2.HttpMethod.GET,
            'POST': tasks_v2.HttpMethod.POST,
        }.get(
            method,
            tasks_v2.HttpMethod.GET
        )

        task = {
            'app_engine_http_request': {
                'app_engine_routing': {'service': service or self.service},
                'headers': headers or {},
                'http_method': http_method,
                'relative_uri': uri,
            }
        }
        if payload:
            task['app_engine_http_request']['body'] = payload

        try:
            self.client.create_task(parent=self.queue.name, task=task, retry=self.retry)
        except GoogleAPIError as e:
            logger.exception(e)



from os import environ, getcwd, path

import invoke

from .. import REGISTRY

BASEPATH = path.basename(getcwd())


@invoke.task(name='build', iterable=['build_arg'])
def build_image(ctx, build_arg=None, name=None, version=None, target=None, test=False):
    """Build the docker image.

    The image name defaults to the name of the working folder.
    It can be overriden with the $SERVICE environment variable.

    The version defaults to 'develop'.
    It can be overriden with the $VERSION environment variable.

    Note that parameters always take precedence over environment variables.
    """
    command = ['docker build']

    if target:
        command.append('--target {}'.format(target))

    if test:
        command.append('--build-arg INSTALL_TEST_DEPENDENCIES=yes')

    for arg in build_arg:
        command.append('--build-arg {}'.format(arg))

    command.append('-t {registry}/{repository}:{version} .'.format(
        registry=REGISTRY,
        repository=name or environ.get('SERVICE', BASEPATH),
        version=version or environ.get('VERSION', 'develop')
    ))

    command = ' '.join(command)
    ctx.run(command)


@invoke.task(name='upload')
def upload_image(ctx, name=None, version=None):
    """Login then upload the docker image to the registry.

    The image name defaults to the name of the working folder.
    It can be overriden with the $SERVICE environment variable.

    The version defaults to 'develop'.
    It can be overriden with the $VERSION environment variable.

    Needless to say, this image must have been already built.

    Note that parameters always take precedence over environment variables.
    """
    ctx.run('`aws ecr get-login --no-include-email`')
    command = "docker push {registry}/{repository}:{version}".format(
        registry=REGISTRY,
        repository=name or environ.get('SERVICE', BASEPATH),
        version=version or environ.get('VERSION', 'develop')
    )
    ctx.run(command)



"""Helpers for retrying functions with exponential back-off.

The `Retry` decorator can be used to retry functions that raise exceptions using exponential
backoff. The retry is limited by a `deadline`. The deadline is the maxmimum amount of time a
method can block.

This decorator will retry transient exceptions.

>>> @retry.Retry()
>>> def call_method():
>>>     return client.method()
"""
import functools
import logging
import random
import time

import arrow

from .. import exceptions

logger = logging.getLogger(__name__)

_DEFAULT_INITIAL_DELAY = 1.0  # seconds
_DEFAULT_MAXIMUM_DELAY = 60.0  # seconds
_DEFAULT_DELAY_MULTIPLIER = 2.0
_DEFAULT_DEADLINE = 60.0 * 2.0  # seconds


def if_exception_type(*exception_types):
    """Creates a predicate to check if the exception is of a given type.

    Parameters:
        exception_types (Sequence[Exception]):
            The exception types to check for.
    Returns:
        Callable[Exception]:
            A predicate that returns True if the provided exception is of the given type(s).
    """

    def if_exception_type_predicate(exception):
        """Bound predicate for checking an exception type."""
        return isinstance(exception, exception_types)

    return if_exception_type_predicate


"""A predicate that checks if an exception is a transient error.

The following server errors are considered transient:
- rbx.exceptions.TransientException
- rbx.exceptions.TransientServerError
"""
if_transient_error = if_exception_type(
    exceptions.TransientException,
    exceptions.TransientServerError,
)


def exponential_sleep_generator(initial, maximum, multiplier=_DEFAULT_DELAY_MULTIPLIER):
    """Generates sleep intervals based on the exponential back-off algorithm.

    This implements the `Truncated Exponential Back-off`_ algorithm:
    https://en.wikipedia.org/wiki/Exponential_backoff

    Parameters:
        initial (float):
            The minimum amount of time to delay. This must be greater than 0.
        maximum (float):
            The maximum amount of time to delay.
        multiplier (float):
            The multiplier applied to the delay.

    Yields:
        float: successive sleep intervals.
    """
    delay = initial
    while True:
        # Introduce jitter by yielding a delay that is uniformly distributed
        # to average out to the delay time.
        yield min(random.uniform(0.0, delay * 2.0), maximum)
        delay = delay * multiplier


def retry_target(target, predicate, sleep_generator, deadline, on_error=None):
    """Call a function and retry if it fails.

    This is the lowest-level retry helper. Generally, you'll use the
    higher-level retry helper :class:`Retry`.

    Parameters:
        target(Callable):
            The function to call and retry.
        predicate (Callable[Exception]):
            A callable used to determine if an exception raised by the target should be considered
            retryable. It should return True to retry or False otherwise.
        sleep_generator (Iterable[float]):
            An infinite iterator that determines how long to sleep between retries.
        deadline (float):
            How long to keep retrying the target. The last sleep period is shortened as necessary,
            so that the last retry runs at ``deadline`` (and not considerably beyond it).
        on_error (Callable[Exception]):
            A function to call while processing a retryable exception. Any error raised by this
            function will *not* be caught.

    Returns:
        Any: the return value of the target function.

    Raises:
        rbx.exceptions.RetryError: If the deadline is exceeded while retrying.
        ValueError: If the sleep generator stops yielding values.
        Exception: If the target raises a method that isn't retryable.
    """
    if deadline is not None:
        deadline_datetime = arrow.utcnow().shift(seconds=deadline).datetime
    else:
        deadline_datetime = None

    last_exc = None

    for sleep in sleep_generator:
        try:
            return target()

        # pylint: disable=broad-except
        # This function explicitly must deal with broad exceptions.
        except Exception as exc:
            if not predicate(exc):
                raise
            last_exc = exc
            if on_error is not None:
                on_error(exc)

        now = arrow.utcnow().datetime

        if deadline_datetime is not None:
            if deadline_datetime <= now:
                raise exceptions.RetryError(
                    f'Deadline of {deadline:.1f}s exceeded while calling {target}',
                    last_exc
                ) from last_exc
            else:
                time_to_deadline = (deadline_datetime - now).total_seconds()
                sleep = min(time_to_deadline, sleep)

        logger.debug(f'Retrying due to {last_exc}, sleeping {sleep:.1f}s ...')
        time.sleep(sleep)

    raise ValueError('Sleep generator stopped yielding sleep values.')


class Retry:
    """Exponential retry decorator.

    This class is a decorator used to add exponential back-off retry behavior.

    Parameters:
        predicate (Callable[Exception]):
            A callable that should return ``True`` if the given exception is retryable.
        initial (float):
            The minimum a,out of time to delay in seconds. This must be greater than 0.
        maximum (float):
            The maximum amout of time to delay in seconds.
        multiplier (float):
            The multiplier applied to the delay.
        deadline (float):
            How long to keep retrying in seconds. The last sleep period is shortened as necessary,
            so that the last retry runs at ``deadline`` (and not considerably beyond it).
    """

    def __init__(
        self,
        predicate=if_transient_error,
        initial=_DEFAULT_INITIAL_DELAY,
        maximum=_DEFAULT_MAXIMUM_DELAY,
        multiplier=_DEFAULT_DELAY_MULTIPLIER,
        deadline=_DEFAULT_DEADLINE,
        on_error=None,
    ):
        self._predicate = predicate
        self._initial = initial
        self._multiplier = multiplier
        self._maximum = maximum
        self._deadline = deadline
        self._on_error = on_error

    def __call__(self, func, on_error=None):
        """Wrap a callable with retry behavior.

        Parameters:
            func (Callable): The callable to add retry behavior to.
            on_error (Callable[Exception]): A function to call while processing
                a retryable exception. Any error raised by this function will
                *not* be caught.

        Returns:
            Callable: A callable that will invoke ``func`` with retry
                behavior.
        """
        if self._on_error is not None:
            on_error = self._on_error

        @functools.wraps(func)
        def retry_wrapped_func(*args, **kwargs):
            """A wrapper that calls target function with retry."""
            target = functools.partial(func, *args, **kwargs)
            sleep_generator = exponential_sleep_generator(
                self._initial, self._maximum, multiplier=self._multiplier
            )
            return retry_target(
                target,
                self._predicate,
                sleep_generator,
                self._deadline,
                on_error=on_error,
            )

        return retry_wrapped_func

    @property
    def deadline(self):
        return self._deadline

    def __str__(self):
        return (
            "<Retry predicate={}, initial={:.1f}, maximum={:.1f}, "
            "multiplier={:.1f}, deadline={:.1f}, on_error={}>".format(
                self._predicate,
                self._initial,
                self._maximum,
                self._multiplier,
                self._deadline,
                self._on_error,
            )
        )



"""Common RBX exceptions."""


class FatalException(Exception):
    """Raise this exception for non-transient errors.

    These exceptions can be given an extra details object, which may be used by the caller to log
    this as extra information.
    """
    message = 'Something went wrong!'

    def __init__(self, message=None, details=None):
        super().__init__(message or self.message)
        self.details = details

    def to_dict(self):
        rv = {
            'message': str(self)
        }

        if self.details:
            rv['details'] = self.details

        return rv


class HTTPException(FatalException):
    """A FatalException with ani optional HTTP status code and a URL."""
    STATUS_CODE = 500

    def __init__(self, message, status_code=None, url=None, **kwargs):
        super().__init__(message, **kwargs)
        self.status_code = status_code or self.STATUS_CODE
        self.url = url

    def to_dict(self):
        rv = super().to_dict()
        rv['status_code'] = self.status_code
        if self.url:
            rv['url'] = self.url

        return rv


class AWSException(Exception):
    """Raised when an unexpected error occurs using boto3."""


class BadRequest(HTTPException):
    """Convenience Exception to use from within a Flask service/blueprint.

    Raising this exception will be caught by the error handler and formatted as an HTTP response,
    using the provided status_code.
    """
    STATUS_CODE = 400


class ClientException(HTTPException):
    """Raised from within the rbx.clients package, caused by third-party APIs.

    This exception is fatal and considered non-transient.
    """


class ClientError(ClientException):
    """4xx client errors."""


class ServerError(ClientException):
    """5xx server errors."""


class TransientServerError(ClientException):
    """Transient 5xx server errors."""


class Unauthorized(ClientException):
    STATUS_CODE = 401


class RetryError(ClientException):
    """Raised when a function has exhausted all of its available retries.

    Parameters:
        message (str):
            The exception message.
        cause (Exception):
            The last exception raised when retring the function.
    """

    def __init__(self, message, cause, status_code=None, url=None):
        super().__init__(message, status_code=status_code, url=url)
        self.message = message
        self.cause = cause

    def __str__(self):
        return f'{self.message}, last exception: {self.cause}'


class TransientException(Exception):
    """Raise this exception when we know the cause is transient (e.g.: connection errors,
    consistency error, ...).
    """



from flask import jsonify

from ..exceptions import BadRequest


def register_error_handlers(app):
    """Register handlers for Exceptions.

    Any exception, whether raised manually or unexpected, will be caught and returned as a
    JSON-formatted response.

    We catch any code-specific Exceptions, namely custom validation (400) and uncaught
    errors (500).

    Custom validation should be raised in the app using the BadRequest class:

    >>> raise BadRequest('Message')

    """

    @app.errorhandler(401)
    def unauthorized(error):
        response = jsonify(message='Unauthorized')
        response.status_code = 401
        return response

    @app.errorhandler(403)
    def forbidden(error):
        response = jsonify(message='Forbidden')
        response.status_code = 403
        return response

    @app.errorhandler(404)
    def not_found(error):
        response = jsonify(message='Not Found')
        response.status_code = 404
        return response

    @app.errorhandler(405)
    def method_not_allowed(error):
        response = jsonify(message='Method Not Allowed')
        response.status_code = 405
        return response

    @app.errorhandler(415)
    def unsupported_media_type(error):
        response = jsonify(message='Unsupported Media Type')
        response.status_code = 415
        return response

    # BadRequest exceptions
    @app.errorhandler(BadRequest)
    def invalid_exception(error):
        app.logger.debug(error.to_dict())
        response = jsonify(error.to_dict())
        response.status_code = error.status_code
        return response

    # Uncaught exceptions
    @app.errorhandler(Exception)
    def unhandled_exception(error):
        app.logger.exception('Unhandled Exception: {}'.format(error))
        response = jsonify(message='Something went wrong!', details=[
            'You have experienced a technical error.',
            'We are working to correct this issue.',
            'Please wait a few moments and try again.',
        ])
        response.status_code = 500
        return response



import invoke


@invoke.task
def check(ctx):
    """
    Check Python package distribution manifest.
    """
    ctx.run('check-manifest -v')


@invoke.task
def build(ctx):
    """
    Build Python package distribution locally.
    """
    ctx.run('python setup.py sdist')


@invoke.task(check, build)
def upload(ctx):
    """
    Check, build, and upload Python package distribution to local PyPI.
    """
    ctx.run('twine upload dist/* -r scoota-pypi')



import logging
import math
import re

from google.api_core.exceptions import Aborted, DeadlineExceeded, ServiceUnavailable
from google.api_core import retry
from google.cloud.datastore import Client, Entity, Key

from ..exceptions import FatalException, TransientException

logger = logging.getLogger(__name__)

# Google API commit object limit per api call
QUERY_LIMIT = 500

default_retry = retry.Retry(
    initial=0.1,
    maximum=60.0,
    multiplier=1.3,
    predicate=retry.if_exception_type(Aborted, DeadlineExceeded, ServiceUnavailable)
)


class Datastore:

    def __init__(self, project, schema=None):
        self.project = project
        self.schema = schema

        self._client = None

    @property
    def client(self):
        if not self._client:
            self._client = Client(project=self.project)
        return self._client

    @property
    def default_retry(self):
        return default_retry

    @property
    def root(self):
        """The root Key for singletons."""
        return self.key('MDM', 'mdm')

    def _get_schema_for_kind(self, kind):
        """Locate and retrieve the schema for the given Kind.

        Raise an FatalException Exception if not found.
        """
        match = [schema for schema in self.schema.values() if schema.name == kind]
        if match:
            return match[0]

        raise FatalException('No schema defined <Kind: {}>'.format(kind))

    def ancestor(self, kind, payload=None, **kwargs):
        """Retrieve an Entity's ancestor.

        Return an Entity.

        The ancestor lookup is extracted from either the payload, or the keyword arguments.

        The lookup field is determined from the schema. When the lookup field is defined in between
        curly brackets, it indicates a field lookup. Otherwise the value is taken as is.
        """
        assert type(self.schema) is dict, 'Schema must be defined for ancestor lookups.'
        schema = self._get_schema_for_kind(kind)

        if not schema.ancestor:
            raise FatalException('No ancestor defined <Schema: {}>'.format(schema))

        return self.fetch_entity(kind=schema.ancestor['kind'],
                                 name=schema.ancestor['lookup'],
                                 stack=payload or kwargs)

    def clone_entity(self, entity):
        """Clone an Entity.

        The clone will be created with an ancestor even when the original wasn't.
        When an ancestor cannot be found, the entity is created without one.

        Note that the clone isn't stored, nor is the original deleted.

        The clone Entity is returned.
        """
        assert type(entity) is Entity

        try:
            parent = self.ancestor(entity.kind, payload=entity)
            key = self.client.key(entity.kind, parent=parent.key)
        except FatalException:
            key = self.client.key(entity.kind)
        except TransientException as e:
            logger.error(f' >> {e} << ')
            return

        clone = Entity(key=key)
        clone.update(entity)

        return clone

    def delete_all_entities_by_kind(self, kind):
        """Delete all Datastore entities by kind."""
        logger.info(f'Deleting "{kind}" from Datastore ')
        query = self.query(kind=kind)
        query.keys_only()

        while True:
            # We can't remove more than QUERY_LIMIT in 1 API call
            keys = tuple(entity.key for entity in query.fetch(
                limit=QUERY_LIMIT))
            if not keys:
                break

            self.client.delete_multi(keys)

    def fetch_checkpoint(self, name):
        """Fetch the Checkpoint entity for the given name."""
        return self.fetch_singleton('Checkpoint', name=name)

    def fetch_entity(self, kind, name, stack=None, key_only=False):
        """Retrieve the Entity or Key from the given kind and name pattern.

        Return an Entity if 'key_only' is False (the default), or a Key
        otherwise.

        When key_only is set as True, only the Key is returned.

        The name pattern may be:

          - '{field_names}': the curly brackets indicate we should attempt to
            find the name value from the stack dictionary.
            The field_names value is a coma-separated list.
            i.e.:
            >>> name = '{field_1,field_2}'
            >>> lookup_value_1 = stack[field_1]
            >>> lookup_value_2 = stack[field_2]

          - 'name': when a single string is provided, it is used as the
            ID/Name of the Key.

        When constructing an Entity/Key via lookup (using the curly brackets
        pattern), we attempt to fetch the Key from Datastore.
        This operation is always expected to find a match, therefore it will
        only return the matching Entity/Key on success.

        It will raise a TransientException on failure.
        """
        results = re.search(r'\{(.*)\}', name)
        if results:
            # The lookup is field-based. The lookup values should be extracted
            # from the stack.
            lookup_fields = results.group(1).split(',')

            if stack is None:
                raise FatalException(
                    'Cannot find a Key via lookup without a stack [ '
                    '<kind: {}>, <name: {}>, <stack: {}>'
                    ' ]'.format(kind, name, stack))

            try:
                lookup_values = [stack[field] for field in lookup_fields]
            except KeyError:
                raise FatalException(
                    'Could not find the lookup fields in the stack [ '
                    '<kind: {}>, <name: {}>, <stack: {}>'
                    ' ]'.format(kind, name, stack))

            query = self.query(kind=kind)
            if key_only:
                query.keys_only()

            for attribute, value in zip(lookup_fields, lookup_values):
                query.add_filter(attribute, '=', value)
            try:
                entity = list(query.fetch())[0]
            except IndexError:
                raise TransientException(
                    'Failed to find the Key [ '
                    '<kind: {}>, <name: {}>, <stack: {}>'
                    ' ]'.format(kind, name, stack))

        else:
            # No lookup required, use the name as the ID/Name.
            entity = Entity(key=self.client.key(kind, name))

        return entity.key if key_only else entity

    def fetch_key(self, *args, **kwargs):
        """A proxy to fetch_entity with the key_only flag set to True."""
        kwargs['key_only'] = True
        return self.fetch_entity(*args, **kwargs)

    def fetch_singleton(self, kind, name, parent=None):
        """Fetch an Entity for the given kind and name.

        When no match is found, return a new Entity.
        Note that this method does not save that new Entity.
        """
        if parent is None:
            parent = self.root

        key = Key(kind, name, parent=parent)

        query = self.query(kind=kind, ancestor=parent)
        query.key_filter(key=key)
        results = list(query.fetch())

        if results:
            return results[0]

        return Entity(key=key)

    def get_day_key(self, date_str):
        """Expects a date string in the form of 'YYYY-MM-DD HH:mm'."""
        try:
            return int(date_str.split(' ')[0].replace('-', ''))
        except ValueError:
            logger.exception('Unable to resolve day_key.')
            return None

    def get_surrogate_key(self, kind, name, entity_id):
        """Get an Entity's surrogate key given its Kind, name, and ID.

        The ID may be a combined ID, in which case it is expected to be a dict instance, keyed by
        the combined IDs.

        Returns None when no match is found.
        """
        key_field = '{}_key'.format(name)

        if isinstance(entity_id, dict):
            id_field = ','.join(f'{field}' for field in entity_id.keys())
            id_field = f'{{{id_field}}}'
            stack = entity_id

        else:
            id_field = f'{{{name}_id}}'
            stack = {f'{name}_id': int(entity_id)}

        try:
            entity = self.fetch_entity(kind, id_field, stack=stack)
            return entity[key_field]
        except TransientException:
            logger.error('Entity <{}: {}> not found in Datastore.'.format(kind, entity_id))

        return None

    def get_time_key(self, date_str):
        """Expects a date string in the form of 'YYYY-MM-DD HH:mm'."""
        try:
            return 10000 + int(date_str.split(' ')[1].replace(':', ''))
        except ValueError:
            logger.exception('Unable to resolve time_key.')
            return None

    def key(self, *args, **kwargs):
        """Proxy to Datastore Client.key()."""
        return self.client.key(*args, **kwargs)

    def make_entity(self, kind, fields=None, parent=None):
        """Create a orphan DS Entity for the given Kind and fields."""
        key = self.client.key(kind, parent=parent or self.root)
        entity = Entity(key)
        if fields:
            entity.update(fields)

        return entity

    def override_entities(self, kind, payload):
        """Given a payload of Entity data, replace all existing Entities for the given Kind."""
        self.delete_all_entities_by_kind(kind)

        entities = [self.make_entity(kind, item) for item in payload]

        self.save(entities)

    def query(self, *args, **kwargs):
        """Proxy to Datastore Client.query()."""
        return self.client.query(*args, **kwargs)

    def relatives(self, kind, payload=None, **kwargs):
        """Retrieve an Entity's relatives.

        The relative lookups are extracted from either the payload, or the keyword arguments.
        The lookup field is determined from the schema.
        """
        assert type(self.schema) is dict, 'Schema must be defined for relatives lookups.'
        schema = self._get_schema_for_kind(kind)

        if type(schema.relatives) not in (list, tuple):
            raise FatalException('No relatives defined <Schema: {}>'.format(schema))

        return [self.fetch_entity(kind=relative['kind'],
                                  name=relative['lookup'],
                                  stack=payload or kwargs)
                for relative in schema.relatives]

    def save(self, entities):
        """Save entities to Datastore.

        Entities are added by batches of QUERY_LIMIT.
        A single Entity instance can be provided, in which case it is saved in
        one single transaction.
        """
        if isinstance(entities, Entity):
            logger.info('Saving 1 record to Datastore.')
            self.client.put(entities, retry=default_retry)
        else:
            logger.info(f'Saving {len(entities)} records to Datastore.')
            saved = 0
            for i in range(0, len(entities), QUERY_LIMIT):
                batch = entities[i:i + QUERY_LIMIT]
                self.client.put_multi(batch, retry=default_retry)
                saved += len(batch)
                logger.debug(f'Saved {saved} of {len(entities)}'
                             f' ({math.floor(100 * saved / len(entities)):d}%).')

    def save_checkpoint(self, checkpoint, timestamp):
        """Set the timestamp for the given Checkpoint Entity.

        The timestamp must be an RFC 3339 timestamp.
        """
        assert type(checkpoint) is Entity

        if checkpoint is not None:
            checkpoint['timestamp'] = timestamp
            self.save(checkpoint)



# Copyright 2021 Rockabox Media Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import find_packages, setup

BOTO = 'boto3==1.17.55'

GOOGLE_DATASTORE = 'google-cloud-datastore>=2.1.0,<2.2'
GOOGLE_LOGGING = 'google-cloud-logging>=2.2.0,<2.3'
GOOGLE_PUBSUB = 'google-cloud-pubsub>=2,<2.5'
GOOGLE_STORAGE = 'google-cloud-storage>=1.36,<2'
GOOGLE_TASKS = 'google-cloud-tasks>=2.0,<3'

AUTH = [
    'google-cloud-firestore>=2.0.2,<2.1'
]
EDS = [
    GOOGLE_DATASTORE,
    'google-cloud-bigquery>=2.9.0,<2.10',
    'rfc3339',
]
MANIFEST = [
    BOTO,
    'opencv-python-headless==4.4.0.46'
]
MAPS = [
    'googlemaps>=4.4.2,<5',
]
MYSQL = [
    'PyMySQL',
    'sqlalchemy>=1.3,<1.4',
]
NOTIFICATIONS = [
    GOOGLE_PUBSUB,
]
QUEUES = [
    'redis==3.5.3',
    'hiredis==2.0.0',
    'rq==1.8.0',
    'rq-scheduler==0.11.0',
]


setup(
    name='rbx',
    version='2.15.1',
    license='Apache 2.0',
    description='Scoota Platform utilities',
    long_description='A collection of common tools for Scoota services.',
    url='http://scoota.com/',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Internet',
    ],
    author='The Scoota Engineering Team',
    author_email='engineering@scoota.com',
    python_requires='>=3.7',
    install_requires=[
        'arrow>=1,<2',
        'Click<8',
        'colorama',
        'PyYAML>=5.4.1',
        'requests>=1.21.1',
    ],
    extras_require={
        # These are requirement bundles required for specific feature sets.
        'auth': AUTH,
        'buildtools': [
            'bumpversion==0.5.3',
            'check-manifest',
            'fabric~=2.5.0',
            'twine',
        ],
        'eds': MYSQL + EDS,
        'geo': MAPS,
        'manifest': MANIFEST,
        'mysql': MYSQL,
        'notifications': NOTIFICATIONS,
        'platform': AUTH + MANIFEST + MAPS + NOTIFICATIONS + QUEUES,
        'queues': QUEUES,

        # These are included for specific libraries. One can either add them directly to their
        # own project, or use one or more of these extras.
        # Their purpose is to show the supported versions of the libraries.
        'aws': [BOTO],
        'datastore': [GOOGLE_DATASTORE],
        'logging': [GOOGLE_LOGGING],
        'pubsub': [GOOGLE_PUBSUB],
        'storage': [GOOGLE_STORAGE],
        'tasks': [GOOGLE_TASKS],

        # Include them all for the test suite.
        'all': AUTH + EDS + MANIFEST + MAPS + MYSQL + NOTIFICATIONS + QUEUES + [
            GOOGLE_LOGGING,
            GOOGLE_STORAGE,
            GOOGLE_TASKS,
        ],
    },
    entry_points={
        'console_scripts': [
            'buildtools = rbx.buildtools.cli:program.run [buildtools]',
            'buildcreative = rbx.manifest.cli:build_creative [manifest]',
            'geocode = rbx.geo.cli:geocode [geo]',
            'reverse_geocode = rbx.geo.cli:reverse_geocode [geo]',
            'unpack = rbx.geo.cli:unpack [geo]',
        ],
    },
    packages=find_packages(),
    zip_safe=True
)



from functools import wraps
import logging
import threading
import traceback
import warnings

from redis import StrictRedis
from rq import Queue as RQ, SimpleWorker, get_current_job
from rq.job import cancel_job

from .scheduler import Scheduler

connection = threading.local()
logger = logging.getLogger(__name__)


class Queue:
    """The Queue provides an interface to farm out jobs to be processed asynchronously."""

    def __init__(self, host='0.0.0.0', port=6379, db=0, module=None):
        """Initialise the Job Queue.

        Parameters:
            host (str):
                The Redis cluster host.
                Uses '0.0.0.0' by default.
            port (int):
                The Redis cluster port.
                Uses port 6379 by default.
            db (int):
                The Redis cluster db.
                Uses 0 by default.
            module:
                The default Queue uses either a method, or as string reference to an importable
                method.
                For string references, this paramater can be used to always load the job method
                from a specific module.

                >>> queue = Queue(module='src.tasks')
                >>> queue.enqueue('method')
                Job(function='src.tasks.method', ...)

        """
        connection.redis = StrictRedis(host=host, port=port, db=db)
        self.connection = connection.redis

        self.queue = RQ(connection=self.connection)
        self.module = module

    @property
    def count(self):
        return self.queue.count

    @property
    def name(self):
        return self.queue.name

    @property
    def exception_handlers(self):
        """The handlers must be listed in order of expected execution."""
        return [self._retry_on_failure,
                self._requeue_job,
                self._move_to_failed_queue]

    @property
    def failed_queue(self):
        return self.queue.failed_job_registry

    def _move_to_failed_queue(self, job, *exc_info):
        """This handler moves the job to the failed queue."""
        exc_string = ''.join(traceback.format_exception(*exc_info))
        logger.debug('Moving job to failed queue')
        self.failed_queue.add(job, ttl=job.failure_ttl, exc_string=exc_string)

        logger.debug(
            'Job {} failed <{}>.'.format(job.id, job.func_name),
            exc_info=exc_info,
            extra={'arguments': job.args}
        )

    def _requeue_job(self, job, *exc_info):
        """This handler does the requeuing, according to the 'failures' meta data
        set by the previous handler.

        When the maximum retry attempts has been reached, we make sure to
        clear the 'failures' flag, and move the job to the next handler in the
        stack (which will move it to the failed queue).
        """
        if 'retry' not in job.meta or job.meta['failures'] >= job.meta['retries']:
            del job.meta['failures']
            job.save()
            return True

        logger.debug('Requeuing {!r}'.format(job))
        self.requeue_job(job)

        return False

    def _retry_on_failure(self, job, *exc_info):
        """This handler flags jobs which are retry-able ('retry' = True) by
        incrementing the 'failures' meta data.

        It will then pass on the handling of the requeuing itself to the next
        handler in the stack.
        If we were to requeue the job at this stage, what would happen is the
        job would be queued up before the next handler had the chance to
        fallback to the next, ending in an infinite loop of failing jobs with
        the same ID.

        The retry() decorator is provided to flag any method as retry-able.

        Also note that scheduled jobs do never get retried, and skip the
        failed queue entirely. They will get retried at the next run anyway.
        """
        if 'scheduled' in job.meta:
            # This is a scheduled job failing. Skip the error handling.
            return False

        job.refresh()
        job.meta.setdefault('failures', 0)
        logger.debug('META: {}'.format(job.meta))

        if 'retry' in job.meta:
            job.meta['failures'] += 1
            logger.debug('Failures: {}'.format(job.meta['failures']))

        return True

    def cancel_job(self, job):
        """Cancel the given job."""
        cancel_job(job.id, connection=self.connection)
        job.delete()

    def cancel_jobs(self, job_ids):
        """Cancel of bunch of jobs given their IDs."""
        for job_id in job_ids:
            self.cancel_job(self.fetch_job(job_id))

    def empty(self):
        """Empty all queues."""
        self.queue.empty()
        for job_id in self.failed_queue.get_job_ids():
            job = self.queue.fetch_job(job_id=job_id)
            self.failed_queue.remove(job)

    def enqueue(self, *args, **kwargs):
        """Register a job to be processed.

        If self.module is set, it is used as the default location for looking up method names.
        The names being the first argument, always.
        """
        if self.module and type(args[0]) is str:
            args = (f'{self.module}.{args[0]}',) + args[1:]

        logger.debug(f'Job queued: {args}, {kwargs}')
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning, module='rq')
            return self.queue.enqueue(*args, **kwargs)

    def fetch_job(self, job_id):
        """Fetch a job for a Job ID."""
        return self.queue.fetch_job(job_id)

    def get_jobs(self, job_ids):
        """Get all jobs."""
        return self.queue.get_jobs()

    def get_worker(self, *args, **kwargs):
        """Return an instance of a worker."""

        class Worker(SimpleWorker):
            def work(self, *args, **kwargs):
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore', category=DeprecationWarning, module='rq')
                    super().work(*args, **kwargs)

        return Worker(
            [self.queue],
            *args,
            connection=self.connection,
            disable_default_exception_handler=True,
            exception_handlers=self.exception_handlers,
            **kwargs
        )

    def requeue_failed_job(self, job_id):
        """Requeue a failed job to its original queue."""
        logger.debug('Requeuing job {}'.format(job_id))
        self.failed_queue.requeue(job_id)

    def requeue_failed_jobs(self, job_ids):
        """Requeue of bunch of failed jobs given their IDs."""
        for job_id in job_ids:
            self.requeue_failed_job(job_id)

    def requeue_job(self, job):
        """Requeue a job that is due to be re-tried after a failure.

        These jobs are pushed back into the queue without having made it to
        the failed queue.

        The requeue_job method provided by the Queue class does not keep the
        meta data, which we use to check how often the job has been tried.
        We therefore requeue these jobs by enqueing the original signature,
        including the meta data.

        Note that jobs retried in this way will be requeued with another ID.
        Therefore the original Job insance will no longer map to the final
        result.
        """
        self.queue.enqueue(job.func, args=job.args, kwargs=job.kwargs, meta=job.meta)


def retry(max_retries=3):
    """Decorator used to move a failing job back into the queue.

    Useful to handle transient errors.
    """
    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            job = get_current_job(connection=connection.redis)
            job.meta['retry'] = True
            job.meta['retries'] = max_retries
            job.save()

            return function(*args, **kwargs)
        return wrapper
    return decorator


__all__ = ['get_current_job', 'Queue', 'retry', 'Scheduler']
